import os
import logging.config
from urllib.parse import urlsplit

from . import defaults, db, buffer as hbuffer, agg, server, metrics, blocks
from .utils import cached_property

TIME_SUFFIXES = {'s': 1, 'm': 60, 'h': 3600, 'd': 86400,
                 'w': 86400 * 7, 'y': 86400 * 365}


def get_config(args):
    result = Config((k, v) for k, v in vars(defaults).items()
                    if not k.startswith('_'))

    config_path = os.environ.get('HISSER_CONFIG') or args.get('_config')
    if config_path:
        from runpy import run_path
        result.update(run_path(config_path))

    for k, v in args.items():
        if not k.startswith('_') and v is not None:
            result[k.upper()] = v

    for k in result:
        ename = 'HISSER_' + k
        if ename in os.environ:
            result[k] = os.environ[ename]

    return result


class Config(dict):
    __getattr__ = dict.__getitem__

    class Error(Exception):
        pass

    def error(func):
        def inner(self, name, *args, **kwargs):
            try:
                return func(self, name, *args, **kwargs)
            except Exception as e:
                raise Config.Error('{}: {}'.format(name, str(e))) from e
        return inner

    @error
    def int(self, name):
        return int(self[name])

    @error
    def float(self, name):
        return float(self[name])

    @error
    def str(self, name):
        param = self[name]
        if not param:
            raise ValueError('Required option')
        return param

    @error
    def host_port(self, name, host='0.0.0.0', port=2003, required=True):
        param = self[name] or ''
        if not param and not required:
            return None
        if param.startswith(':'):
            param = host + param
        url = urlsplit('tcp://' + param)
        return url.hostname, url.port or port

    def ensure_dirs(self):
        blocks.ensure_block_dirs(self.data_dir, self.retentions)

    @cached_property
    def data_dir(self):
        return self.str('DATA_DIR')

    @cached_property
    def retentions(self):
        return parse_retentions(self['RETENTIONS'])

    @cached_property
    def agg_rules(self):
        default = self.str('AGG_DEFAULT_METHOD')
        return agg.AggRules(get_agg_rules_from_dict(self), default)

    @cached_property
    def storage(self):
        return db.Storage(data_dir=self.data_dir,
                          retentions=self.retentions,
                          merge_finder=self.merge_finder,
                          downsample_finder=self.downsample_finder,
                          agg_rules=self.agg_rules,
                          metric_index=self.metric_index)

    @cached_property
    def block_list(self):
        return db.BlockList(data_dir=self.data_dir)

    @cached_property
    def merge_finder(self):
        max_size = self.int('MERGE_MAX_SIZE')
        max_gap_size = self.int('MERGE_MAX_GAP_SIZE')
        ratio = self.float('MERGE_RATIO')

        def finder(resolution, blocks):  # pragma: nocover
            return db.find_blocks_to_merge(resolution, blocks, max_size=max_size,
                                           max_gap_size=max_gap_size, ratio=ratio)
        return finder

    @cached_property
    def downsample_finder(self):
        max_size = self.int('DOWNSAMPLE_MAX_SIZE')
        min_size = self.int('DOWNSAMPLE_MIN_SIZE')
        max_gap_size = self.int('MERGE_MAX_GAP_SIZE')

        def finder(resolution, blocks, new_resolution, start=0):  # pragma: nocover
            return db.find_blocks_to_downsample(
                resolution, blocks, new_resolution,
                max_size=max_size, min_size=min_size,
                max_gap_size=max_gap_size, start=start
            )

        return finder

    @cached_property
    def buffer(self):
        min_res = self.retentions[0][0]
        return hbuffer.Buffer(size=self.int('BUFFER_SIZE'),
                              resolution=min_res,
                              flush_size=self.int('BUFFER_FLUSH_SIZE'),
                              past_size=self.int('BUFFER_PAST_SIZE'),
                              max_points=self.int('BUFFER_MAX_POINTS'),
                              compact_ratio=self.float('BUFFER_COMPACT_RATIO'))

    @cached_property
    def reader(self):
        return db.Reader(block_list=self.block_list,
                         retentions=self.retentions,
                         rpc_client=self.rpc_client)

    @cached_property
    def server(self):
        return server.Server(
            buf=self.buffer,
            storage=self.storage,
            carbon_host_port_tcp=self.host_port('CARBON_BIND'),
            carbon_host_port_udp=self.host_port('CARBON_BIND_UDP', required=False),
            link_host_port=self.host_port('LINK_BIND', required=False),
            backlog=self.int('CARBON_BACKLOG')
        )

    @cached_property
    def rpc_client(self):
        host_port = self.host_port('LINK_BIND', required=False)
        if host_port:
            return server.RpcClient(host_port)

    @cached_property
    def metric_index(self):
        return metrics.MetricIndex(os.path.join(self.data_dir, 'metric.index'))

    def setup_logging(self, daemon=True):  # pragma: nocover
        if daemon and self.LOGGING:
            logging.config.dictConfig(self.LOGGING)
        else:
            logging.basicConfig(level=self.LOGGING_LEVEL,
                                format='[%(asctime)s] %(name)s:%(levelname)s %(message)s')


def parse_retentions(string):
    result = (part.split(':') for part in string.split(','))
    return sorted((parse_seconds(res), parse_seconds(ret)) for res, ret in result)


def parse_aggregation(string):
    return string.rsplit('|', 1)


def parse_seconds(interval):
    if isinstance(interval, int):
        return interval

    interval = interval.strip()
    if interval.isdigit():
        return int(interval)

    return int(interval[:-1]) * TIME_SUFFIXES[interval[-1]]


def get_agg_rules_from_dict(cfg):
    rules = sorted((k, v)
                   for k, v in cfg.items()
                   if k.startswith('AGG_RULE_') and v)

    return [parse_aggregation(v) for _, v in rules]



# Default parameters for hisser
#
# All time values can have interval suffixes: (s)econds, (m)inutes,
# (h)ours, (d)ays, (w)eeks and (y)ears. Seconds by default.

# Directory with data, can be set via `hisser` `--data-dir` option.
DATA_DIR = None

# Default retentions, resolution:keep-interval,...
# Greater resolution must be an integer factor of a smaller one.
RETENTIONS = '1m:3d, 5m:15d, 30m:90d, 2h:1y'

# Default aggregations, pattern|method,...
# Patterns are applied in sorted order.
# You can test it with `hisser agg-method <metric name>` command.
AGG_RULE_MAX = r'\.max$|max'
AGG_RULE_MIN = r'\.min$|min'
AGG_RULE_SUM = r'\.count$|sum'
# AGG_RULE_CUSTOM_RULE = r'my_pattern|avg'

# Default aggregation if no any rules matched
AGG_DEFAULT_METHOD = 'avg'

# How much points of each metric to keep in memory.
# With default minimal resolution of 60s it will be 30 minutes of data.
BUFFER_SIZE = 30

# How much data can be accepted from the past. 5 * 60s is 5 minutes.
BUFFER_PAST_SIZE = 5

# Flush buffer on reaching this size. Buffer will be flushed every 10 minutes
# with default flush size and 60s resolution.
BUFFER_FLUSH_SIZE = 10

# Flush buffer on reaching points limit. Points are equal to number of metrics in
# buffer multiplied by current buffer position. For example if your instance
# receive 100k metrics per minute, then it will be flushed every 5 minutes for default
# resolution of 60s.
BUFFER_MAX_POINTS = 500000

BUFFER_COMPACT_RATIO = 0.9

MERGE_MAX_SIZE = 700
MERGE_MAX_GAP_SIZE = 30
MERGE_RATIO = 1.4

DOWNSAMPLE_MIN_SIZE = 10
DOWNSAMPLE_MAX_SIZE = 1000

# Listen tcp `[host]:port` for carbon text protocol,
# by default host is 0.0.0.0.
CARBON_BIND = ':2003'

# Listen udp `[host]:port` for carbon text protocol
CARBON_BIND_UDP = None

# Listen backlog for carbon protocol
CARBON_BACKLOG = 100

# Listen tcp `[host]:port` for link protocol,
LINK_BIND = '127.0.0.1:8002'

# Listen backlog for link protocol
LINK_BACKLOG = 100

# Python logging dict, if none log to stdout
LOGGING = None

# Default log level
LOGGING_LEVEL = 'ERROR'



import re
import logging
from contextlib import contextmanager
from time import clock, perf_counter

from graphite.node import LeafNode, BranchNode
from graphite.finders.utils import BaseFinder
from graphite.render.grammar import grammar

from . import config

log = logging.getLogger('hisser.graphite')


@contextmanager
def profile(name):  # pragma: no cover
    real = perf_counter()
    cpu = clock()
    try:
        yield
    finally:
        real_duration = perf_counter() - real
        cpu_duration = clock() - cpu
        print('##', name, real_duration, cpu_duration, flush=True)


def scream(fn):  # pragma: nocover
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception:
            log.exception('Error getting data from hisser')
            raise
    return inner


def parse_tagspec(tagspec):
    m = re.match('^([^;!=]+)(!?=~?)([^;]*)$', tagspec)
    if m is None:  # pragma: no cover
        raise ValueError("Invalid tagspec %s" % tagspec)

    tag = m.group(1)
    operator = m.group(2)
    spec = m.group(3)

    return tag, operator, spec


class Finder(BaseFinder):
    tags = True

    def __init__(self, cfg=None):
        self.cfg = cfg or config.get_config({})
        self.reader = self.cfg.reader
        self.metric_index = self.cfg.metric_index

    @scream
    def find_nodes(self, query):
        for l, r in self.metric_index.find_tree(query.pattern):
            if l:
                yield LeafNode(r.decode(), None)
            else:
                yield BranchNode(r.decode())

    def find_metric_by_tags_many(self, patterns, cache):
        result = {}
        for pattern in patterns:
            args = grammar.parseString(pattern).expression.call.args
            exprs = [t.string[1:-1] for t in args if t.string]
            queries = [parse_tagspec(e) for e in exprs]
            result[pattern] = self.metric_index.match_by_tags(queries, cache)
        return result

    @scream
    def fetch(self, patterns, start_time, stop_time, now=None, requestContext=None):
        spatterns = []
        tpatterns = []
        cache = {}

        for pattern in patterns:
            if pattern.startswith('seriesByTag('):
                tpatterns.append(pattern)
            else:
                spatterns.append(pattern)

        names = set()
        smetrics = tmetrics = None

        if spatterns:
            smetrics = self.metric_index.find_metrics_many(spatterns)
            for v in smetrics.values():
                names.update(v)

        if tpatterns:
            tmetrics = self.find_metric_by_tags_many(tpatterns, cache)
            for v in tmetrics.values():
                names.update(v)

        names = sorted(names)
        time_info, data = self.reader.fetch(names, int(start_time), int(stop_time))

        result = []

        for metrics in filter(None, (smetrics, tmetrics)):
            for query, names in metrics.items():
                for name in names:
                    values = data.get(name)
                    if values:
                        result.append({
                            'pathExpression': query,
                            'path': name.decode(),
                            'name': name.decode(),
                            'time_info': time_info,
                            'values': values,
                        })

        return result

    @scream
    def auto_complete_tags(self, exprs, tagPrefix=None, limit=None, requestContext=None):
        result = tags = [r.decode() for r in self.metric_index.get_tags()]
        if tagPrefix:
            result = [r for r in result if r.startswith(tagPrefix)]
            if len(result) < 10:
                rset = set(result)
                result.extend(r for r in tags
                              if tagPrefix in r and r not in rset)
        return result

    @scream
    def auto_complete_values(self, exprs, tag, valuePrefix=None, limit=None, requestContext=None):
        result = values = [r.decode() for r in self.metric_index.get_tag_values(tag)]
        if valuePrefix:
            result = [r for r in result if r.startswith(valuePrefix)]
            if len(result) < 10:
                rset = set(result)
                result.extend(r for r in values
                              if valuePrefix in r and r not in rset)
        return result



import os.path
import pathlib
from collections import namedtuple

try:
    from os import scandir
except ImportError:  # pragma: nocover
    from scandir import scandir


class Block(namedtuple('Block', 'start end idx size resolution path')):
    @staticmethod
    def make(start, size, resolution, path):
        return Block(start, start + size * resolution, 0, size, resolution, path)

    def split(self, ts):
        return Slice.make(self).split(ts)

    def slice(self, start, stop=None):
        return Slice.make(self).slice(start, stop)


class Slice(namedtuple('Slice', 'start end idx size resolution path bstart')):
    @staticmethod
    def make(block):
        return Slice(block.start, block.end, 0, block.size,
                     block.resolution, block.path, block.start)

    def split(self, ts):
        if ts <= self.start:
            return None, self
        elif ts >= self.end:
            return self, None

        return self.slice_to(ts), self.slice_from(ts)

    def slice(self, start, stop=None):
        result = self
        if start is not None:
            result = result.slice_from(start)
        if stop is not None:
            result = result and result.slice_to(stop)
        return result

    def slice_from(self, ts):
        if ts <= self.start:
            return self

        if ts >= self.end:
            return None

        start = ts
        end = self.end
        return self._replace(start=start, end=end,
                             idx=(start - self.bstart) // self.resolution,
                             size=(end - start) // self.resolution)

    def slice_to(self, ts):
        if ts <= self.start:
            return None

        if ts >= self.end:
            return self

        start = self.start
        end = ts
        return self._replace(start=start, end=end,
                             idx=(start - self.bstart) // self.resolution,
                             size=(end - start) // self.resolution)


class BlockList:
    def __init__(self, data_dir):
        self.data_dir = data_dir
        self._last_state = {}
        self._blocks = {}

    def check(self, resolution, refresh):
        if refresh or resolution not in self._last_state:
            self.rescan(resolution)
            self._last_state[resolution] = 0
            return

        try:
            new_state = os.path.getmtime(block_state_name(self.data_dir, resolution))
        except OSError:
            new_state = 0

        if self._last_state[resolution] < new_state:
            self._last_state[resolution] = new_state
            self.rescan(resolution)

    def blocks(self, resolution, refresh=False):
        self.check(resolution, refresh)
        return self._blocks[resolution]

    def rescan(self, resolution):
        blocks = self._blocks[resolution] = []
        data_path = os.path.join(self.data_dir, str(resolution))

        try:
            entries = scandir(data_path)
        except FileNotFoundError:
            os.makedirs(data_path, exist_ok=True)
            entries = []

        for e in entries:
            if e.name.endswith('.hdb') and e.is_file():
                try:
                    info = get_info(e.path, resolution)
                except ValueError:
                    pass
                else:
                    blocks.append(info)

        blocks.sort()


def get_info(path, res=0):
    ts, size, *rest = os.path.basename(path).split('.')
    ts, size = int(ts), int(size)
    return Block.make(ts, size, res, path)


def block_state_name(data_dir, resolution):
    return os.path.join(data_dir, str(resolution), 'blocks.state')


def notify_blocks_changed(data_dir, resolution):
    pathlib.Path(block_state_name(data_dir, resolution)).touch(exist_ok=True)


def ensure_block_dirs(data_dir, retentions):
    for r, _ in retentions:
        os.makedirs(os.path.join(data_dir, str(r)), exist_ok=True)



import re
from math import isnan

from .utils import NAN


def is_not_nan(num):
    return not isnan(num)


def _sum(data):
    non_empty = list(filter(is_not_nan, data))
    return sum(non_empty), len(non_empty)


def safe_avg(data):
    total, n = _sum(data)
    if n:
        return total / n
    return NAN


def safe_sum(data):
    total, n = _sum(data)
    if n:
        return total
    return NAN


def safe_max(data):
    return max(filter(is_not_nan, data), default=NAN)


def safe_min(data):
    return min(filter(is_not_nan, data), default=NAN)


def safe_last(data):
    try:
        return list(filter(is_not_nan, data))[-1]
    except IndexError:
        return NAN


METHODS = {
    'avg': safe_avg,
    'sum': safe_sum,
    'max': safe_max,
    'min': safe_min,
    'last': safe_last
}


class AggRules:
    def __init__(self, rules, default='avg'):
        self.rules = tuple((re.compile(r), METHODS[m]) for r, m in rules)
        self.rules_bin = tuple((re.compile(r.encode()), METHODS[m]) for r, m in rules)
        self.default = METHODS[default]

    def get_methods(self, names, use_bin=False):
        result = {}
        rules = self.rules_bin if use_bin else self.rules
        for n in names:
            for p, v in rules:
                if p.search(n):
                    result[n] = v
                    break
        return result, self.default

    def get_method(self, name, use_bin=False):
        rules = self.rules_bin if use_bin else self.rules
        for p, v in rules:
            if p.search(name):
                return v
        return self.default



import os
import time
import errno
import socket
import signal
import fcntl
import logging
from selectors import DefaultSelector, EVENT_READ, EVENT_WRITE

from .utils import run_in_fork, wait_childs, mloads, mdumps

log = logging.getLogger(__name__)


def rpc_fetch(srv, keys):
    return srv.buf.get_data(keys)


class Server:
    def __init__(self, buf, storage, carbon_host_port_tcp,
                 carbon_host_port_udp=None, link_host_port=None, backlog=100):
        self.buf = buf
        self.storage = storage
        self.carbon_host_port_tcp = carbon_host_port_tcp
        self.carbon_host_port_udp = carbon_host_port_udp
        self.link_host_port = link_host_port
        self.backlog = backlog

        self.ready_to_merge = False
        self.flush_pids = set()
        self.merge_pid = None

        self.time_to_exit = False

    def accept(self, sock, cdata):
        conn, _addr = sock.accept()
        conn.setblocking(False)
        self.sel.register(conn, EVENT_READ, (cdata['handler'], {}))

    def carbon_read_tcp(self, conn, cdata):
        data = conn.recv(4096)
        olddata = cdata.get('buf', b'')
        if data:
            cdata['buf'] = self.process(olddata + data)
        else:
            if olddata:
                self.process(olddata, True)
            self.sel.unregister(conn)
            conn.close()

    def carbon_read_udp(self, conn, cdata):
        (data, _addr) = conn.recvfrom(4096)
        if data:
            self.process(data, end=True)

    def link_write(self, conn, cdata):
        buf = cdata['buf']
        count = conn.send(buf)
        cdata['buf'] = buf = buf[count:]

        if not buf:
            self.sel.unregister(conn)
            conn.close()

    def link_read(self, conn, cdata):
        data = conn.recv(4096)
        buf = cdata.get('buf', b'')
        if data:
            cdata['buf'] = buf + data
            return

        if not buf:  # pragma: nocover
            self.sel.unregister(conn)
            conn.close()
            return

        try:
            req = mloads(buf)
            method = req.pop('method')
            resp = mdumps(globals()['rpc_{}'.format(method)](self, **req))
        except Exception as e:
            resp = mdumps({'error': str(e)})

        self.sel.unregister(conn)
        self.sel.register(conn, EVENT_WRITE, (self.link_write, {'buf': resp}))

    def process(self, data, end=False):
        lines = data.splitlines(True)
        next_chunk = b''
        if not end and not lines[-1].endswith(b'\n'):
            next_chunk = lines[-1]
            lines = lines[:-1]

        buf = self.buf
        for line in lines:
            parts = line.split()

            try:
                name = parts[0]
                value = float(parts[1])
                ts = int(float(parts[2]))
            except (ValueError, IndexError):
                pass
            else:
                buf.add(ts, name, value)

        return next_chunk

    def signal_read(self, conn, cdata):  # pragma: nocover
        data = os.read(conn, 4096)
        if data:
            if data[-1] in (signal.SIGINT, signal.SIGTERM):
                log.info('Cought exit signal')
                self.time_to_exit = True

    def setup_signals(self, sel):  # pragma: nocover
        self.pipe_r, pipe_w = os.pipe()
        flags = fcntl.fcntl(pipe_w, fcntl.F_GETFL, 0)
        flags = flags | os.O_NONBLOCK
        fcntl.fcntl(pipe_w, fcntl.F_SETFL, flags)
        signal.set_wakeup_fd(pipe_w)

        sel.register(self.pipe_r, EVENT_READ, (self.signal_read, None))

        def dummy(signal, frame):
            pass

        signal.signal(signal.SIGINT, dummy)
        signal.signal(signal.SIGTERM, dummy)

    def listen(self, signals=True):
        sel = self.sel = DefaultSelector()

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(self.carbon_host_port_tcp)
        sock.listen(self.backlog)
        sock.setblocking(False)
        sel.register(sock, EVENT_READ, (self.accept, {'handler': self.carbon_read_tcp}))

        if self.carbon_host_port_udp:
            sock_udp = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock_udp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock_udp.bind(self.carbon_host_port_udp)
            sock_udp.setblocking(False)
            sel.register(sock_udp, EVENT_READ, (self.carbon_read_udp, None))

        if self.link_host_port:
            link_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            link_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            link_sock.bind(self.link_host_port)
            link_sock.listen(self.backlog)
            link_sock.setblocking(False)
            sel.register(link_sock, EVENT_READ, (self.accept, {'handler': self.link_read}))

        if signals:  # pragma: nocover
            self.setup_signals(sel)

    def check_childs(self):
        if self.flush_pids or self.merge_pid:
            try:
                pid, _exit = wait_childs()
            except OSError as e:  # pragma: nocover
                if e.errno == errno.ECHILD:
                    self.flush_pids.clear()
                    self.merge_pid = None
                else:
                    raise
            else:
                if self.flush_pids and pid in self.flush_pids:
                    self.flush_pids.remove(pid)
                    self.ready_to_merge = True
                if self.merge_pid and self.merge_pid == pid:
                    self.merge_pid = None
            return True
        return False

    def check_loop(self):
        events = self.sel.select(3)
        for key, _mask in events:
            callback, data = key.data
            callback(key.fileobj, data)

    def check_buffer(self, now=None):
        data, new_names = self.buf.tick(now=now)
        if data:
            self.flush_pids.add(run_in_fork(self.storage.new_block, *data).pid)
            self.ready_to_merge = False

        if new_names:
            self.flush_pids.add(run_in_fork(self.storage.new_names, new_names).pid)

        if self.ready_to_merge and not self.merge_pid:
            self.merge_pid = run_in_fork(self.storage.do_housework).pid
            self.ready_to_merge = False

    def run(self):
        while not self.time_to_exit:
            self.check_loop()
            self.check_childs()
            self.check_buffer()

        while self.check_childs():  # pragma: nocover
            time.sleep(1)

        data, _new_names = self.buf.tick(force=True)
        if data:
            self.storage.new_block(*data)


class RpcClient:
    def __init__(self, host_port=('127.0.0.1', 8002), connect_timeout=1, timeout=5):
        self.host_port = host_port
        self.connect_timeout = connect_timeout
        self.timeout = timeout

    def call(self, method, **kwargs):
        s = socket.create_connection(self.host_port, self.connect_timeout)
        s.settimeout(self.timeout)
        kwargs['method'] = method
        s.sendall(mdumps(kwargs))
        s.shutdown(socket.SHUT_WR)
        payload = b''
        while True:
            data = s.recv(4096)
            if not data:
                break
            payload += data

        s.close()
        return mloads(payload)



import os
import resource
import array

from time import time
from functools import partial
from math import ceil
from collections import namedtuple
from contextlib import contextmanager

import msgpack
import lmdb

NAN = float('nan')
MB = 1 << 20
PAGE_SIZE = resource.getpagesize()

mdumps = partial(msgpack.dumps, use_bin_type=True)
mloads = partial(msgpack.loads, encoding='utf-8')

Fork = namedtuple('Fork', 'pid start')


def norm_res(ts, res):
    return int(ts) // res * res


def estimate_data_size(data, size):
    return (1000 * len(data) + size * 8 * len(data))


def page_size(size):
    return max(PAGE_SIZE, ceil(size / PAGE_SIZE) * PAGE_SIZE)


def safe_unlink(path):
    try:
        os.unlink(path)
    except OSError:
        pass


def map_size_for_path(path):
    return page_size(os.path.getsize(path))


def overlap(range1, range2):  # pragma: nocover
    left = max(range1[0], range2[0])
    right = min(range1[1], range2[1])
    result = right - left
    if result > 0:
        return result, left, right
    else:
        return 0, 0, 0


def run_in_fork(func, *args, **kwargs):
    pid = os.fork()
    if pid == 0:  # pragma: nocover
        try:
            func(*args, **kwargs)
        except Exception:
            import traceback
            traceback.print_exc()
            os._exit(1)
        else:
            os._exit(0)
    else:
        return Fork(pid, time())


def wait_childs():
    return os.waitpid(-1, os.WNOHANG)


class cached_property(object):
    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:  # pragma: nocover
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


def open_env(path, map_size=None, readonly=False, lock=None):
    lock = not readonly if lock is None else lock
    try:
        if map_size and map_size < 0:
            map_size = map_size_for_path(path) - map_size
        else:
            map_size = map_size or map_size_for_path(path)
    except FileNotFoundError:
        map_size = 10*MB
    return lmdb.open(path, map_size, subdir=False, readonly=readonly, lock=lock)


@contextmanager
def cursor(path, map_size=None, readonly=False, lock=None, buffers=False):
    with open_env(path, map_size, readonly=readonly, lock=lock) as env:
        with env.begin(write=not readonly, buffers=buffers) as txn:
            with txn.cursor() as cur:
                yield cur


@contextmanager
def txn_cursor(env, write=False, db=None):
    with env.begin(write=write) as txn:
        with txn.cursor(db) as cur:
            yield cur


def empty_rows(data, size):
    nanbuf = array.array('d', [NAN] * size).tobytes()
    for k, v in data:
        if v.tobytes() == nanbuf:
            yield k


def non_empty_rows(data, size):
    nanbuf = array.array('d', [NAN] * size).tobytes()
    for k, v in data:
        if v.tobytes() != nanbuf:
            yield k, v



import logging
import os.path
import heapq

from math import isnan
from time import time
from itertools import islice, groupby

from .blocks import Block, BlockList, notify_blocks_changed, get_info
from .utils import (estimate_data_size, mdumps, mloads, NAN, safe_unlink,
                    MB, page_size, norm_res, cursor, non_empty_rows, open_env)

log = logging.getLogger(__name__)


def abs_ratio(a, b):
    return max(a, b) / (min(a, b) or 1)


class Reader:
    def __init__(self, block_list, retentions, rpc_client):
        self.block_list = block_list
        self.retentions = retentions
        self.rpc_client = rpc_client

    def fetch(self, names, start, stop, res=None, rest_res=None):
        if not res:
            resolutions = [r[0] for r in self.retentions]
            resolutions.reverse()
            res = min(resolutions, key=lambda r: abs_ratio((stop - start) // r, 1000))
            rest_res = [r for r in resolutions if r < res]

        blocks = self.block_list.blocks(res)
        start = start // res * res
        stop = rstop = stop // res * res + res

        result = {}
        blocks = [b for b in blocks if b.end > start and b.start < stop]
        if blocks:
            blocks[0] = blocks[0].slice(start, stop)
            blocks[-1] = blocks[-1].slice(start, stop)

            start = blocks[0].start
            size = (blocks[-1].end - start) // res
            empty_row = [None] * size
            for b in blocks:
                r_start_idx = (b.start - start) // res
                r_end_idx = r_start_idx + b.size
                c_end_idx = b.idx + b.size
                data = read_block(b.path, names)
                for k, values in data.items():
                    try:
                        row = result[k]
                    except KeyError:
                        row = result[k] = empty_row[:]

                    row[r_start_idx:r_end_idx] = [None if isnan(v) else v
                                                  for v in values[b.idx:c_end_idx]]

            stop = start + size * res
        else:
            stop = start
            size = 0

        if not rest_res and rstop > stop:
            return self.add_rest_data_from_buffer(names, start, stop, rstop, res, size, result)
        return (start, stop, res), result

    def add_rest_data_from_buffer(self, keys, start, stop, rstop, res, size, result):
        if not self.rpc_client:
            return (start, stop, res), result

        try:
            cur_data = self.rpc_client.call('fetch', keys=keys)
        except Exception:
            log.exception('Error getting data')
            return (start, stop, res), result

        cur_result = cur_data['result']
        cur_slice = Block.make(cur_data['start'], cur_data['size'],
                               cur_data['resolution'], 'tmp')
        ib = cur_slice.slice(stop, rstop)
        if ib:
            add = [None] * ((ib.end - stop) // res)
            s_idx = size + (ib.start - stop) // res
            for name in keys:
                row = result.get(name)
                if row is None:
                    row = result[name] = [None] * size + add
                else:
                    row += add
                values = cur_result.get(name)
                if values is not None:
                    row[s_idx: s_idx + ib.size] = values[ib.idx:ib.idx+ib.size]
            stop = ib.end

        return (start, stop, res), result


class Storage:
    def __init__(self, data_dir, retentions, merge_finder, downsample_finder,
                 agg_rules, metric_index):
        self.data_dir = data_dir
        self.retentions = retentions
        self.merge_finder = merge_finder
        self.downsample_finder = downsample_finder
        self.agg_rules = agg_rules
        self.metric_index = metric_index

    def new_block(self, data, ts, resolution, size, new_names):
        self.new_names(new_names)
        data = sorted((k, list(v)) for k, v in non_empty_rows(data, size))
        if data:
            return new_block(self.data_dir, data, ts, resolution, size, append=True)

    def new_names(self, new_names):
        if new_names:
            self.metric_index.add(sorted(new_names))

    def do_housework(self, now=None):
        self.do_merge()
        self.do_downsample()
        self.do_cleanup(now)

    def do_merge(self):
        block_list = BlockList(self.data_dir)
        for res, _ in self.retentions:
            blocks = block_list.blocks(res)
            for s in self.merge_finder(res, blocks):
                log.info('Merge %r', s)
                merge(self.data_dir, res, s)

    def do_downsample(self):
        block_list = BlockList(self.data_dir)
        resolutions = [r[0] for r in self.retentions]
        for res, new_res in zip(resolutions[:-1], resolutions[1:]):
            blocks = block_list.blocks(res)
            if not blocks:
                continue
            new_blocks = block_list.blocks(new_res)
            start = new_blocks and new_blocks[-1].end or 0
            segments = self.downsample_finder(res, blocks, new_res, start)
            if segments:
                downsample(self.data_dir, new_res, segments, self.agg_rules)

    def do_cleanup(self, now=None):
        block_list = BlockList(self.data_dir)
        now = now or time()
        for res, ret in self.retentions:
            for b in block_list.blocks(res):
                if b.end < now - ret:
                    os.unlink(b.path)
                    log.info('Cleanup old block %s', b.path)
            notify_blocks_changed(self.data_dir, res)


def split_descending_blocks(blocks, ratio):
    blocks = blocks[::-1]
    for idx, (p, n) in enumerate(zip(blocks[:-1], blocks[1:]), 1):
        if p.size / n.size > ratio:
            break
    else:
        idx = None

    if idx is not None:
        r1 = blocks[idx:][::-1]
        r2 = blocks[:idx][::-1]
    else:
        r1 = []
        r2 = blocks[::-1]

    if len(r2) > 1:
        for p, n in zip(r2[:-1], r2[1:]):
            if max(p.size, n.size) / min(p.size, n.size) <= ratio:
                r2 = [p, n]
                break
        else:
            r2 = []

    return [r1, r2]


def find_blocks_to_merge(resolution, blocks, *, max_size, max_gap_size, ratio):
    result = []
    segment = []
    it = iter(blocks)
    b = None
    while True:
        b = b or next(it, None)
        if not b:
            break

        if not segment:
            sstart = send = b.start
        else:
            sstart = segment[0].start
            send = segment[-1].end

        is_ok = True
        if is_ok and b.start - send > max_gap_size * resolution:
            is_ok = False

        if is_ok and (b.end - sstart) // resolution > max_size:
            is_ok = False

        if is_ok:
            segment.append(b)
            b = None
        elif segment:
            result.append(segment)
            segment = []

    if segment:
        result.append(segment)

    if result and len(result[-1]) > 1:
        result = result[:-1] + split_descending_blocks(result[-1], ratio)

    return [[b.path for b in s] for s in result if len(s) > 1]


def find_blocks_to_downsample(resolution, blocks, new_resolution,
                              max_gap_size, min_size, max_size, start=0):
    assert new_resolution % resolution == 0
    start = norm_res(start, new_resolution)
    result = []
    segment = None
    it = (r for r in blocks if r.end > start)
    b = None
    while True:
        b = b or next(it, None)
        if not b:
            break

        prev = segment and segment[-1]
        if not segment or (b.start - prev.end) // new_resolution > max_gap_size:
            segment = []
            if b.start <= start:
                s_start = start
            else:
                s_start = norm_res(b.start, new_resolution)
            stop = norm_res(s_start + max_size * resolution, new_resolution)
            result.append((segment, s_start))

        bs = b.slice(s_start)
        if not bs:  # pragma: nocover
            break
        cur, b = bs.split(stop)
        s_start = cur.end
        segment.append(cur)
        if s_start >= stop:
            segment = []

    if result:
        last = result[-1][0]
        ssize = (last[-1].end - last[0].start) // new_resolution
        if ssize < min_size:
            result = result[:-1]

    final_result = []
    for segment, s_start in result:
        s_stop = norm_res(segment[-1].end, new_resolution)
        if s_stop < segment[-1].end:
            s_stop += new_resolution
        final_result.append((segment, s_start, s_stop))

    return final_result


def downsample(data_dir, new_resolution, segments, agg_rules):
    for (blocks, s_start, s_stop) in segments:
        iters = [iter_dump(b.path, idx) for idx, b in enumerate(blocks)]
        stream = groupby(heapq.merge(*iters), lambda r: r[0])

        resolution = blocks[0].resolution
        s_size = (s_stop - s_start) // resolution
        f_size = (s_stop - s_start) // new_resolution
        csize = new_resolution // resolution
        empty_row = [NAN] * s_size
        max_size, max_block = max((os.path.getsize(b.path), b) for b in blocks)
        map_size = page_size(max_size * f_size / max_block.size * 5)

        s_slices = []
        b_slices = []
        for b in blocks:
            idx = (b.start - s_start) // resolution
            s_slices.append(slice(idx, idx+b.size))
            b_slices.append(slice(b.idx, b.idx+b.size))

        def gen():
            for k, g in stream:
                row = empty_row[:]
                for _, bn, values in g:
                    row[s_slices[bn]] = values[b_slices[bn]]

                agg_method = agg_rules.get_method(k, use_bin=True)
                agg = [agg_method(row[r:r+csize]) for r in range(0, s_size, csize)]
                yield k, agg

        path = new_block(data_dir, gen(), s_start, new_resolution, s_size // csize,
                         map_size=map_size, append=True)
        log.info('Downsample %s', path)


def merge(data_dir, res, paths):
    blocks = [get_info(p, res) for p in paths]
    iters = [iter_dump(b.path, idx) for idx, b in enumerate(blocks)]

    first = blocks[0]
    last = blocks[-1]
    size = (last.end - first.start) // res
    empty_row = [NAN] * size

    max_size, max_block = max((os.path.getsize(b.path), b) for b in blocks)
    map_size = page_size(max_size * size / max_block.size * 3)

    slices = []
    overlaps = []
    last_idx = None
    for b in blocks:
        idx = (b.start - first.start) // res
        slices.append(slice(idx, idx+b.size))
        overlaps.append(last_idx and idx <= last_idx)
        last_idx = max(last_idx or 0, idx + b.size)

    stream = groupby(heapq.merge(*iters), lambda r: r[0])

    def gen():
        for k, g in stream:
            row = empty_row[:]
            for _, bn, values in g:
                if overlaps[bn]:
                    values = [r if isnan(v) else v
                              for r, v in zip(row[slices[bn]], values)]
                row[slices[bn]] = values
            yield k, row

    new_block(data_dir, gen(), first.start, res, size,
              map_size=map_size, append=True, notify=False)

    for p in paths:
        os.unlink(p)
        safe_unlink(p + '-lock')

    notify_blocks_changed(data_dir, res)


def new_block(data_dir, data, timestamp, resolution, size,
              map_size=None, append=False, notify=True):
    fname = '{}.{}.hdb'.format(timestamp, size)
    path = os.path.join(data_dir, str(resolution), fname)
    tmp_path = path + '.tmp'

    map_size = map_size or estimate_data_size(data, size) * 2 + 100*MB
    data = ((k, mdumps(v)) for k, v in data)
    with cursor(tmp_path, page_size(map_size), lock=False) as cur:
        cur.putmulti(data, overwrite=False, append=append)

    os.rename(tmp_path, path)

    if notify:
        notify_blocks_changed(data_dir, resolution)

    return path


def read_block(path, keys):
    result = {}
    with cursor(path, readonly=True) as cur:
        for k in keys:
            v = cur.get(k, None)
            if v is not None:
                result[k] = mloads(v)
    return result


def dump(path):  # pragma: nocover
    with cursor(path, readonly=True) as cur:
        for k, v in cur:
            yield k, mloads(v)


def iter_dump(path, idx, size=10000):
    k = None
    while True:
        with open_env(path, readonly=True) as env:
            with env.begin(write=False) as txn:
                with txn.cursor() as cur:
                    if k:
                        cur.set_key(k)
                    else:
                        cur.first()

                    for k, v in islice(cur, size):
                        yield k, idx, mloads(v)

                    if cur.next():
                        k = cur.key()
                    else:
                        break



import logging
from math import isnan
from time import time
from array import array
from resource import getrusage, RUSAGE_SELF, RUSAGE_CHILDREN

from .utils import NAN, empty_rows

log = logging.getLogger(__name__)


class Buffer:
    def __init__(self, size, resolution, flush_size, past_size, max_points,
                 compact_ratio, now=None):
        self.size = size
        self.resolution = resolution
        self.flush_size = flush_size
        self.past_size = past_size
        self.max_points = max_points
        self.compact_ratio = compact_ratio

        self.data = {}
        self.new_names = []
        self.names_to_check = []
        self.collected_metrics = 0

        self.empty_row = array('d', [NAN] * size)
        self.past_points = 0
        self.future_points = 0
        self.received_points = 0
        self.flushed_points = 0
        self.last_size = 0

        self.set_ts(now or time())

    def get_data(self, keys):
        result = {}
        for k in keys:
            try:
                result[k] = list(self.data[k])
            except KeyError:
                pass

        return {'start': self.ts,
                'result': result,
                'resolution': self.resolution,
                'size': self.size}

    def cut_data(self, size):
        result = []
        empty_part = array('d', [NAN] * size)
        for k, v in self.data.items():
            result.append((k, v[:size]))
            v[:-size] = v[size:]
            v[-size:] = empty_part[:]
        return result

    def get_row(self, name):
        try:
            return self.data[name]
        except KeyError:
            self.new_names.append(name)
        result = self.data[name] = self.empty_row[:]
        return result

    def set_ts(self, ts):
        self.ts = int(ts) // self.resolution * self.resolution - self.past_size * self.resolution

    def flush(self, size):
        self.flushed_points += len(self.data) * size

        data = self.cut_data(size)
        if data:
            result = (data, self.ts, self.resolution, size, self.new_names[:])

            estimated_metrics = self.collected_metrics // size
            if not self.names_to_check and estimated_metrics / len(self.data) < self.compact_ratio:
                log.info('Compact data %d -> %d', len(self.data), estimated_metrics)
                self.names_to_check = list(self.data)
        else:
            result = None

        self.ts += self.resolution * size
        self.collected_metrics = 0
        self.last_size = 0
        self.new_names[:] = []
        return result

    def add(self, ts, name, value):
        self.received_points += 1
        idx = (int(ts) - self.ts) // self.resolution
        try:
            row = self.get_row(name)
            oldvalue = row[idx]
            row[idx] = value
            self.collected_metrics += isnan(oldvalue)
        except IndexError:
            if idx < 0:
                self.past_points += 1
            else:
                self.future_points += 1

    def check_and_drop_names(self):
        if not self.names_to_check:
            return

        log.info('Check for empty names %d', len(self.names_to_check))
        names = self.names_to_check[-10000:]
        self.names_to_check = self.names_to_check[:-10000]
        d = self.data
        for n in empty_rows(((k, d[k]) for k in names), self.size):
            del d[n]

    def add_internal_metrics(self, now):
        self.add(now, b'hisser.flushed-points', self.flushed_points)
        self.add(now, b'hisser.received-points', self.received_points)
        self.add(now, b'hisser.past-points', self.past_points)
        self.add(now, b'hisser.future-points', self.future_points)

        r_main = getrusage(RUSAGE_SELF)
        self.add(now, b'hisser.cpu.main.user', r_main.ru_utime)
        self.add(now, b'hisser.cpu.main.sys', r_main.ru_stime)
        self.add(now, b'hisser.mem.main.maxrss', r_main.ru_maxrss)
        self.add(now, b'hisser.io.main.blocks_read', r_main.ru_inblock)
        self.add(now, b'hisser.io.main.blocks_write', r_main.ru_oublock)

        r_forks = getrusage(RUSAGE_CHILDREN)
        self.add(now, b'hisser.cpu.forks.user', r_forks.ru_utime)
        self.add(now, b'hisser.cpu.forks.sys', r_forks.ru_stime)
        self.add(now, b'hisser.mem.forks.maxrss', r_forks.ru_maxrss)
        self.add(now, b'hisser.io.forks.blocks_read', r_forks.ru_inblock)
        self.add(now, b'hisser.io.forks.blocks_write', r_forks.ru_oublock)

    def tick(self, force=False, now=None):
        self.check_and_drop_names()

        now = int(now or time())
        size = (now - self.past_size * self.resolution - self.ts) // self.resolution

        if size < 0:
            return None, None

        if size != self.last_size:
            self.add_internal_metrics(now)
            self.last_size = size

        if force:
            size = (now - self.ts) // self.resolution
            return self.flush(min(size, self.size)), None

        if size >= self.size:
            return self.flush(self.size), None

        if size >= self.flush_size:
            return self.flush(self.flush_size), None

        if size * len(self.data) > self.max_points:
            return self.flush(size), None

        if size > 0 and self.new_names:
            new_names = self.new_names[:]
            self.new_names[:] = []
            return None, new_names

        return None, None



from setuptools import setup, find_packages
import hisser

setup(
    name='hisser',
    version=hisser.version,
    url='https://github.com/baverman/hisser',
    author='Anton Bobrov',
    author_email='baverman@gmail.com',
    license='MIT',
    description='Fast TSDB backend for graphite',
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=['tests']),
    install_requires=['msgpack', 'click', 'lmdb'],
    entry_points={
        'console_scripts': ['hisser = hisser.cli:cli']
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)



import re
from fnmatch import filter as fnfilter, translate

import lmdb
from .utils import MB, txn_cursor


class MetricIndex:
    def __init__(self, path, map_size=500*MB):
        self.env = lmdb.open(path, map_size, subdir=False, max_dbs=4)
        self.tree_db = self.env.open_db(b'tree', dupsort=True)
        self.tag_name_db = self.env.open_db(b'tag_name', dupsort=True)
        self.tag_db = self.env.open_db(b'tag', dupsort=True)

    def add(self, names):
        tagged = [r for r in names if b';' in r]
        simple = [r for r in names if b';' not in r]

        if simple:
            tree = make_tree(simple)
            with txn_cursor(self.env, True, self.tree_db) as cur:
                cur.putmulti(tree)

        if tagged:
            tags, tag_names = parse_tags(tagged)
            with txn_cursor(self.env, True, self.tag_db) as cur:
                cur.putmulti(tags)
            with txn_cursor(self.env, True, self.tag_name_db) as cur:
                cur.putmulti(tag_names)

    def iter_tree(self):
        with txn_cursor(self.env, False, self.tree_db) as cur:
            for k, v in cur:
                yield k, v

    def find_metrics_many(self, queries, check=False):
        matched_metrics = {}
        with txn_cursor(self.env, False, self.tree_db) as cur:
            for q in queries:
                prefix, parts = query_parts(q)
                candidates = [prefix or b'.']
                for pattern in parts:
                    to_match = {}
                    for c in candidates:
                        if cur.set_key(c):
                            prefix = b'' if c == b'.' else c + b'.'
                            for m in cur.iternext_dup():
                                to_match.setdefault(m, []).append(prefix + m)

                    candidates[:] = []
                    for m in fnfilter(to_match, pattern):
                        candidates.extend(to_match[m])

                if check:
                    matched_metrics[q] = [(not cur.set_key(c), c) for c in sorted(candidates)]
                else:
                    matched_metrics[q] = sorted(candidates)
            return matched_metrics

    def find_metrics(self, query):
        return self.find_metrics_many([query]).get(query, [])

    def find_tree(self, query):
        return self.find_metrics_many([query], True).get(query, [])

    def find_by_tag_values(self, tag, values):
        result = []
        with txn_cursor(self.env, False, self.tag_name_db) as cur:
            for value in values:
                key = '{}={}'.format(tag, value).encode()
                if cur.set_key(key):
                    result.extend(cur.iternext_dup(False, True))
        return result

    def get_tags(self):
        with txn_cursor(self.env, False, self.tag_db) as cur:
            return list(cur.iternext_nodup(True, False))

    def get_tag_values(self, tag):
        with txn_cursor(self.env, False, self.tag_db) as cur:
            if cur.set_key(tag.encode()):
                return list(cur.iternext_dup(False, True))
        return []

    def match_by_tags(self, queries, cache=None):
        if cache is None:
            cache = {}

        result = None
        for tag, op, value in queries:
            if op == '=':
                values = [value]
            elif op == '!=':
                values = [r for r in self.cached_tag_values(tag, cache)
                          if r != value]
            elif op == '=~':
                values = pattern_match(self.cached_tag_values(tag, cache), value)
            elif op == '!=~':
                values = pattern_not_match(self.cached_tag_values(tag, cache), value)
            else:  # pragma: no cover
                continue

            names = self.find_by_tag_values(tag, values)
            if result is None:
                result = set(names)
            else:
                result.intersection_update(names)

            if not result:
                return set()

        return result

    def cached_tag_values(self, tag, cache):
        try:
            return cache[tag]
        except KeyError:
            pass
        result = cache[tag] = [r.decode() for r in self.get_tag_values(tag)]
        return result


def make_tree(names):
    empty_row = [None] * 255
    prev = empty_row[:]
    for n in names:
        prefix = None
        parts = n.split(b'.')
        for idx, (pp, p) in enumerate(zip(prev, parts)):
            if pp != p:
                prev[idx] = p
                prev[idx+1:255] = empty_row[:255-idx-1]
                yield (prefix or b'.', p)
            if prefix:
                prefix += b'.' + p
            else:
                prefix = p


def query_parts(query):
    parts = query.encode().split(b'.')
    prefix = []
    for q in parts:
        if any(r in q for r in (b'*', b'[', b']')):
            break
        else:
            prefix.append(q)
    return b'.'.join(prefix), parts[len(prefix):]


def parse_tags(names):
    tags = []
    tag_names = []
    for name in names:
        parts = name.split(b';')
        if not parts:  # pragma: no cover
            continue

        tags.append((b'name', parts[0]))
        tag_names.append((b'name=%b' % parts[0], name))

        for p in parts[1:]:
            k, _, v = p.partition(b'=')
            tags.append((k, v))
            tag_names.append((b'%b=%b' % (k, v), name))

    return tags, tag_names


def pattern_match(values, pattern):
    if pattern.startswith(':'):
        enum = set(filter(None, pattern[1:].split(',')))
        return [r for r in values if r in enum]
    elif pattern.startswith('!'):
        m = re.compile(translate(pattern[1:])).match
        return [r for r in values if m(r)]
    else:
        m = re.compile(pattern).match
        return [r for r in values if m(r)]


def pattern_not_match(values, pattern):
    if pattern.startswith(':'):
        enum = set(filter(None, pattern[1:].split(',')))
        return [r for r in values if r not in enum]
    elif pattern.startswith('!'):
        m = re.compile(translate(pattern[1:])).match
        return [r for r in values if not m(r)]
    else:
        m = re.compile(pattern).match
        return [r for r in values if not m(r)]



import sys
from functools import wraps

import click
from . import config, db, defaults, agg, metrics, utils, version


@click.group()
@click.version_option(version=version)
def cli():
    pass


def config_aware(func):
    @wraps(func)
    def inner(**kwargs):
        cfg = config.get_config(kwargs)
        cfg.setup_logging(func.__name__ == 'cmd_run')
        try:
            return func(cfg)
        except config.Config.Error as e:
            print(str(e), file=sys.stderr)
            sys.exit(1)

    return inner


def common_options(func):
    func = click.option('--config', '-c', '_config',
                        metavar='path', help='path to config file')(func)
    func = click.option('--data-dir', '-d', metavar='path',
                        help='path to directory with data')(func)
    return func


@cli.command('merge', help='merge two or more blocks')
@common_options
@click.option('-r', 'resolution', type=int)
@click.argument('block', nargs=-1)
@config_aware
def cmd_merge(cfg):
    if cfg.BLOCK:
        db.merge(cfg.data_dir, cfg.RESOLUTION, list(cfg.BLOCK))
    else:
        cfg.storage.do_merge()


@cli.command('downsample', help='run downsample')
@common_options
@config_aware
def cmd_downsample(cfg):
    cfg.storage.do_downsample()


@cli.command('cleanup', help='remove old blocks')
@common_options
@config_aware
def cmd_cleanup(cfg):
    cfg.storage.do_cleanup()


@cli.command('dump', help='dump content of block')
@click.argument('block')
@config_aware
def cmd_dump(cfg):
    for k, v in db.dump(cfg.BLOCK):
        print(k, len(v), v)


@cli.command('dump-index', help='dump content of metric index')
@click.argument('index')
@config_aware
def cmd_dump_index(cfg):
    mi = metrics.MetricIndex(cfg.INDEX)
    for k, v in mi.iter_tree():
        print(k.decode(), v.decode())


@cli.command('backup', help='backup db file')
@click.argument('dbfile')
@click.argument('out')
@config_aware
def cmd_backup(cfg):
    with open(cfg.OUT, 'wb') as f:
        with utils.open_env(cfg.DBFILE) as env:
            env.copyfd(f.fileno(), True)


@cli.command('check', help='checks metadata')
@click.argument('block', nargs=-1)
@config_aware
def cmd_check(cfg):
    for path in cfg.BLOCK:
        block = db.get_info(path)
        result = set()
        for _k, v in db.dump(path):
            result.add(len(v))

        if len(result) > 2 or list(result)[0] != block.size:
            print(path, 'Invalid sizes', sorted(result))


@cli.command('run', help='run server')
@common_options
@click.option('--carbon-bind', '-l', metavar='[host]:port',
              help=('host and port to listen carbon text'
                    ' protocol on tcp, default is {}').format(defaults.CARBON_BIND))
@click.option('--carbon-bind-udp', metavar='[host]:port',
              help=('host and port to listen carbon'
                    ' text protocol on udp, default is {}').format(defaults.CARBON_BIND_UDP))
@click.option('--link-bind', metavar='[host]:port',
              help=('host and port to listen graphite finder link protocol'
                    ', default is {}').format(defaults.LINK_BIND))
@config_aware
def cmd_run(cfg):
    cfg.ensure_dirs()
    server = cfg.server
    server.listen()
    server.run()


@cli.command('agg-method', help='show aggregation method for metric')
@click.argument('metric', metavar='metric.name')
@config_aware
def cmd_agg_method(cfg):
    method = cfg.agg_rules.get_method(cfg.METRIC)
    rmethods = {v: k for k, v in agg.METHODS.items()}
    print(rmethods[method])


if __name__ == '__main__':
    cli(prog_name='hisser')
